# -*- coding: utf-8 -*-
###########################################################################################
#   googleapi - connect to google api through oauth 2
#
#   Date        Author      Reason
#   ----        ------      ------
#   12/02/17    Lou King    adapted from https://developers.google.com/identity/protocols/OAuth2WebServer
#   02/15/19    Lou King    updated due to demise of google+ api
#
#   Copyright 2017 Lou King
###########################################################################################

# standard
import os.path

#pypi
from flask import jsonify, url_for, session, request, abort
from flask.views import View
import httplib2
from oauth2client import client

# import google.oauth2.credentials
# import google_auth_oauthlib.flow

# note as of this writing, oauth2client is deprecated. 
# See https://github.com/GoogleCloudPlatform/google-auth-library-python/blob/master/docs/oauth2client-deprecation.rst
# but there is no support in the replacement lib for Storage, and google claims they will maintain without adding features
# so this is probably ok
from oauth2client.file import Storage

############################################################################
class GoogleAuth(View):
############################################################################

    #----------------------------------------------------------------------
    def __init__( self, app, client_secrets_file, scopes, startendpoint, credfolder=None, 
                  logincallback=lambda email: None, logoutcallback=lambda email: None,
                  loginfo=None, logdebug=None, logerror=None, ):
    #----------------------------------------------------------------------
        '''
        :param app: flask application
        :param client_secrets_file: client_secrets.json path
        :param scopes: list of google scopes. see https://developers.google.com/identity/protocols/googlescopes
        :param startendpoint: endpoint to start with after authorization completed (no leading slash)
        :param credfolder: folder where credential Storage will be placed
        :param logincallback: function(email) called when login detected
        :param logoutcallback: function called when logout detected
        :param loginfo: info logger function
        :param logdebug: debug logger function
        :param logerror: debug logger function
        '''
        self.app = app
        self.client_secrets_file = client_secrets_file
        self.scopes = scopes
        self.startendpoint = startendpoint
        self.credfolder = credfolder
        self.logincallback = logincallback
        self.logoutcallback = logoutcallback
        self.loginfo = loginfo
        self.logdebug = logdebug
        self.logerror = logerror

        # create supported endpoints
        # _token is ajax call from google sign-in
        self.app.add_url_rule('/_token', view_func=self.token, methods=['POST',])
        self.app.add_url_rule('/clear', view_func=self.clear_credentials, methods=['GET',])

    #----------------------------------------------------------------------
    def token(self):
    #----------------------------------------------------------------------
        # (Receive auth_code by HTTPS POST)
        auth_code = request.data

        # If this request does not have `X-Requested-With` header, this could be a CSRF
        if not request.headers.get('X-Requested-With'):
            abort(403)

        # Exchange auth code for access token, refresh token, and ID token
        credentials = client.credentials_from_clientsecrets_and_code(
            self.client_secrets_file,
            self.scopes,
            auth_code)

        # Get profile info from ID token
        user_id = credentials.id_token['sub']
        email = credentials.id_token['email']

        authorized = False
        if user_id:
            credfile = os.path.join(self.credfolder, user_id)
            storage = Storage(credfile)
            # do the new credentials not have a refresh token? do we already have credentials? 
            # if so, just update the access token and expiry from new credentials into stored and resave
            storedcred = storage.get()
            if not credentials.refresh_token and storedcred:
                storedcred.access_token = credentials.access_token
                storedcred.token_expiry = credentials.token_expiry
                credentials = storedcred
            credentials.set_store(storage)
            storage.put(credentials)
            session['_ga_google_user_id'] = user_id
            session['_ga_google_email'] = email
            # refresh
            http = httplib2.Http()
            credentials.refresh(http)

            # take care of login specifics
            authorized = self.logincallback(email)

        if self.logdebug: self.logdebug( 'oauth2callback() session = {}'.format(session) )

        return jsonify( {'authorized':authorized, 'redirect':url_for(self.startendpoint)} )

    #----------------------------------------------------------------------
    def clear_credentials(self):
    #----------------------------------------------------------------------
        if 'credentials' in session:
            del session['credentials']
        if '_ga_google_user_id' in session:
            del session['_ga_google_user_id']
        
        email = None
        if '_ga_google_email' in session:
            email = session['_ga_google_email']
            del session['_ga_google_email']

        # take care of logout specifics
        self.logoutcallback(email)

        return 'Credentials have been cleared from session cookie'

    #----------------------------------------------------------------------
    def get_userid(self, credentials):
    #----------------------------------------------------------------------
        try:
            user_id = session['_ga_google_user_id']
            if self.logdebug: self.logdebug( 'get_userid() retrieved user_id from session cookie' )

        except KeyError:
            try:
                user_id = credentials.id_token['sub']
                if self.logdebug: self.logdebug( 'get_userid() profile from credentials = {}'.format(credentials) )
            
            except:
                if self.logdebug: self.logdebug( 'invalid credentials, continuing' )
                user_id = None

        return user_id

#----------------------------------------------------------------------
def get_credentials(credfolder):
#----------------------------------------------------------------------
    try:
        user_id = session['_ga_google_user_id']
        credfile = os.path.join(credfolder, user_id)
        storage = Storage(credfile)
        credentials = storage.get()
        credentials.set_store(storage)

    except KeyError:
        credentials = None

    return credentials

#----------------------------------------------------------------------
def get_email():
#----------------------------------------------------------------------
    try:
        email = session['_ga_google_email']

    except KeyError:
        email = None

    return email




#!/usr/bin/python
###########################################################################################
# renderrun - common functions for rendering information related to running
#
#	Date		Author		Reason
#	----		------		------
#       02/24/13        Lou King        Create
#       03/07/14        Lou King        Copied from runningclub
#
#   Copyright 2013,2014 Lou King
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###########################################################################################
'''
renderrun - common functions for rendering information related to running
==============================================================================

'''

# standard
import math

# pypi

# github

# other

# home grown
from . import timeu

dbtime = timeu.asctime('%Y-%m-%d')
rndrtim = timeu.asctime('%m/%d/%Y')

class softwareError(Exception): pass

#----------------------------------------------------------------------
def getprecision(distance,surface='road'): 
#----------------------------------------------------------------------
    '''
    get the precision for rendering, based on distance
    
    precision might be different for time vs. age group adjusted time
    
    :param distance: distance (miles)
    :param surface: 'road', 'track' or 'trail', default 'road'
    :rtype: (timeprecision,agtimeprecision)
    '''
    
    meterspermile = 1609    # close enough, and this is the value used in agegrade.py
    
    # See rule 165 from http://www.usatf.org/usatf/files/f1/f101f17f-1b8b-4f1b-ac9c-d9022155aa94.pdf
    # all hand timed track races have resolution to 0.1 seconds; all road races have resolution to 1 seconds
    # all times are rounded up (done by caller of this method)
    # TODO: allow configuration of timing method (hand timed vs automatically timed)

    # track events
    if surface=='track':
        timeprecision = 1
        agtimeprecision = 1

    # road and trail events
    else:
        timeprecision = 0
        agtimeprecision = 0

    return timeprecision, agtimeprecision

#----------------------------------------------------------------------
def renderdate(dbdate): 
#----------------------------------------------------------------------
    '''
    create date for display
    
    :param dbdate: date from database ('yyyy-mm-dd')
    '''
    try:
        dtdate = dbtime.asc2dt(dbdate)
        rval = rndrtim.dt2asc(dtdate)
    except ValueError:
        rval = dbdate
    return rval

#----------------------------------------------------------------------
def adjusttime(rawtime,precision,useceiling=True): 
#----------------------------------------------------------------------
    '''
    adjust raw time based on precision
    
    :param rawtime: time in seconds
    :param precision: number of places after decimal point
    :param useceiling: True if ceiling function to be used (round up)
    
    :rtype: adjusted time in seconds (float)
    '''
    # shift time based on precision
    multiplier = 10**precision

    # multiply whole time by multiplier to get integral time
    # then take ceiling or round
    # then divide by multiplier to get whole and fractional part
    fixedtime = rawtime * multiplier
    if useceiling:
        adjfixedtime = math.ceil(fixedtime)
    else:
        adjfixedtime = round(fixedtime)
    adjtime = adjfixedtime / multiplier
    
    return adjtime

#----------------------------------------------------------------------
def rendertime(dbtime,precision,useceiling=True): 
#----------------------------------------------------------------------
    '''
    create time for display
    
    :param dbtime: time in seconds
    :param precision: number of places after decimal point
    :param useceiling: True if ceiling function to be used (round up)
    '''
    
    if precision > 0:
        ''' old code
        multiplier = 10**precision
        # note round up per USATF rule 165
        fracdbtime = dbtime - int(dbtime)
        if useceiling:
            frac = int(math.ceil(fracdbtime*multiplier))
        else:
            frac = int(round(fracdbtime*multiplier))
        if frac < multiplier:
            rettime = fracformat.format(frac)
            remdbtime = int(dbtime)
        else:
            rettime = fracformat.format(0)
            remdbtime = int(dbtime+1)
        '''
        
        # adjust time based on precision
        adjtime = adjusttime(dbtime,precision,useceiling)
        
        # update the rendering what will be returned to include fractional part and what remains
        wholetime = int(adjtime)
        fractime = adjtime - wholetime
        fracformat = '{{0:0.{0}f}}'.format(precision)
        rettime = fracformat.format(fractime)
        remdbtime = wholetime
        
        # retttime should have leading 0.  remove it
        if rettime[0] != '0':
            raise softwareError('formatted adjusted time fraction does not have leading 0: {0}'.format(adjtime))
        rettime = rettime[1:]
        
    else:
        # note round up per USATF rule 165
        if useceiling:
            remdbtime = int(math.ceil(dbtime))
        else:
            remdbtime = int(round(dbtime))
        rettime = ''
    
    thisunit = remdbtime%60
    firstthru = True
    while remdbtime > 0:
        if not firstthru:
            rettime = ':' + rettime
        firstthru = False
        rettime = '{0:02d}'.format(thisunit) + rettime
        remdbtime /= 60
        thisunit = remdbtime%60
        
    while rettime[0] == '0':
        rettime = rettime[1:]
        
    return rettime



###########################################################################################
# userrole - manage application users and roles
#
#       Date            Author          Reason
#       ----            ------          ------
#       12/08/19        Lou King        Create
#
#   Copyright 2019 Lou King
#
###########################################################################################
'''
userrole - manage application users and roles
====================================================
'''

# standard

# pypi
from validators.slug import slug
from validators.email import email
from flask import g
from flask_security.recoverable import send_reset_password_instructions

# homegrown
from . import bp
from loutilities.user.model import db, User, Role, Interest, Application
from loutilities.tables import DbCrudApiRolePermissions, get_request_action, SEPARATOR

##########################################################################################
# users endpoint
###########################################################################################

user_dbattrs = 'id,email,name,given_name,roles,interests,last_login_at,current_login_at,last_login_ip,current_login_ip,login_count,active'.split(',')
user_formfields = 'rowid,email,name,given_name,roles,interests,last_login_at,current_login_at,last_login_ip,current_login_ip,login_count,active'.split(',')
user_dbmapping = dict(zip(user_dbattrs, user_formfields))
user_formmapping = dict(zip(user_formfields, user_dbattrs))

def user_validate(action, formdata):
    results = []

    if formdata['email'] and not email(formdata['email']):
        results.append({ 'name' : 'email', 'status' : 'invalid email: correct format is like john.doe@example.com' })

    # check apps which user will have access to
    apps = set()
    if formdata['roles'] and 'id' in formdata['roles'] and formdata['roles']['id'] != '':
        roleidsstring = formdata['roles']['id']
        roleids = roleidsstring.split(SEPARATOR)
        for roleid in roleids:
            thisrole = Role.query.filter_by(id=roleid).one()
            apps |= set(thisrole.applications)

    # this app must be one of user's roles
    if g.loutility not in apps:
        # need to use name='roles.id' because this field is _treatment:{relationship}
        results.append({'name': 'roles.id', 'status': 'give user at least one role which works for this application'})

    return results

class UserCrudApi(DbCrudApiRolePermissions):
    def editor_method_posthook(self, form):
        '''
        send new users a link to set their password

        :param form: edit form
        :return: None
        '''
        action = get_request_action(form)
        if action == 'create':
            user = User.query.filter_by(id=self.created_id).one()
            send_reset_password_instructions(user)

class UserView(UserCrudApi):
    def __init__(self, **kwargs):
        '''
        application MUST instantiate UserView

        application should override editor_method_postcommit(self, form) to call
        loutilities.model.ManageLocalUser(db, appname, localusermodel, localinterestmodel).update()
        '''
        self.kwargs = kwargs
        args = dict(
            app=bp,  # use blueprint instead of app
            db=db,
            model=User,
            version_id_col='version_id',  # optimistic concurrency control
            roles_accepted='super-admin',
            template='datatables.jinja2',
            pagename='users',
            endpoint='userrole.users',
            rule='/users',
            dbmapping=user_dbmapping,
            formmapping=user_formmapping,
            clientcolumns=[
                {'data': 'email', 'name': 'email', 'label': 'Email', '_unique': True,
                 'className': 'field_req',
                 },
                {'data': 'given_name', 'name': 'given_name', 'label': 'First Name',
                 'className': 'field_req',
                 },
                {'data': 'name', 'name': 'name', 'label': 'Full Name',
                 'className': 'field_req',
                 },
                {'data': 'roles', 'name': 'roles', 'label': 'Roles',
                 '_treatment': {'relationship': {'fieldmodel': Role, 'labelfield': 'name', 'formfield': 'roles',
                                                 'dbfield': 'roles', 'uselist': True}}
                 },
                {'data': 'interests', 'name': 'interests', 'label': 'Interests',
                 '_treatment': {'relationship': {'fieldmodel': Interest, 'labelfield': 'description',
                                                 'formfield': 'interests', 'dbfield': 'interests',
                                                 'uselist': True}}
                 },
                {'data': 'active', 'name': 'active', 'label': 'Active',
                 '_treatment': {'boolean': {'formfield': 'active', 'dbfield': 'active'}},
                 'ed': {'def': 'yes'},
                 },
                {'data': 'last_login_at', 'name': 'last_login_at', 'label': 'Last Login At', 'type': 'readonly'},
                {'data': 'current_login_at', 'name': 'current_login_at', 'label': 'Current Login At',
                 'type': 'readonly'},
                {'data': 'last_login_ip', 'name': 'last_login_ip', 'label': 'Last Login IP', 'type': 'readonly'},
                {'data': 'current_login_ip', 'name': 'current_login_ip', 'label': 'Current Login IP',
                 'type': 'readonly'},
                {'data': 'login_count', 'name': 'login_count', 'label': 'Login Count', 'type': 'readonly'},
            ],
            validate=user_validate,
            servercolumns=None,  # not server side
            idSrc='rowid',
            buttons=['create', 'editRefresh'],
            dtoptions={
                'scrollCollapse': True,
                'scrollX': True,
                'scrollXInner': "100%",
                'scrollY': True,
            },
        )
        args.update(kwargs)
        super().__init__(**args)


##########################################################################################
# roles endpoint
###########################################################################################

role_dbattrs = 'id,name,description,applications'.split(',')
role_formfields = 'rowid,name,description,applications'.split(',')
role_dbmapping = dict(zip(role_dbattrs, role_formfields))
role_formmapping = dict(zip(role_formfields, role_dbattrs))

class RoleView(DbCrudApiRolePermissions):
    def __init__(self, **kwargs):
        '''
        application MUST instantiate RoleView

        application should override editor_method_postcommit(self, form) to call
        loutilities.model.ManageLocalUser(db, appname, localusermodel, localinterestmodel).update()
        '''
        self.kwargs = kwargs
        args = dict(
                    app = bp,   # use blueprint instead of app
                    db = db,
                    model = Role, 
                    version_id_col = 'version_id',  # optimistic concurrency control
                    roles_accepted = 'super-admin',
                    template = 'datatables.jinja2',
                    pagename = 'roles', 
                    endpoint = 'userrole.roles',
                    rule = '/roles',
                    dbmapping = role_dbmapping, 
                    formmapping = role_formmapping, 
                    clientcolumns = [
                        { 'data': 'name', 'name': 'name', 'label': 'Name',
                          'className': 'field_req',
                          },
                        { 'data': 'description', 'name': 'description', 'label': 'Description' },
                        {'data': 'applications', 'name': 'applications', 'label': 'Applications',
                         '_treatment': {'relationship': {'fieldmodel': Application, 'labelfield': 'application',
                                                         'formfield': 'applications', 'dbfield': 'applications',
                                                         'uselist': True}}
                         },
                    ],
                    servercolumns = None,  # not server side
                    idSrc = 'rowid', 
                    buttons = ['create', 'editRefresh', 'remove'],
                    dtoptions = {
                                        'scrollCollapse': True,
                                        'scrollX': True,
                                        'scrollXInner': "100%",
                                        'scrollY': True,
                                  },
                    )
        args.update(kwargs)
        super().__init__(**args)

##########################################################################################
# interests endpoint
###########################################################################################

interest_dbattrs = 'id,interest,description,users,public,applications'.split(',')
interest_formfields = 'rowid,interest,description,users,public,applications'.split(',')
interest_dbmapping = dict(zip(interest_dbattrs, interest_formfields))
interest_formmapping = dict(zip(interest_formfields, interest_dbattrs))

def interest_validate(action, formdata):
    results = []

    for field in ['interest']:
        if formdata[field] and not slug(formdata[field]):
            results.append({ 'name' : field, 'status' : 'invalid slug: must be only alpha, numeral, hyphen' })

    return results

class InterestView(DbCrudApiRolePermissions):
    def __init__(self, **kwargs):
        '''
        application MUST instantiate InterestView

        application should override editor_method_postcommit(self, form) to call
        loutilities.model.ManageLocalUser(db, appname, localusermodel, localinterestmodel).update()
        '''
        self.kwargs = kwargs
        args = dict(
            app=bp,  # use blueprint instead of app
            db=db,
            model=Interest,
            version_id_col='version_id',  # optimistic concurrency control
            interests_accepted='super-admin',
            template='datatables.jinja2',
            pagename='interests',
            endpoint='userrole.interests',
            rule='/interests',
            dbmapping=interest_dbmapping,
            formmapping=interest_formmapping,
            clientcolumns=[
                {'data': 'description', 'name': 'description', 'label': 'Description', '_unique': True,
                 'className': 'field_req',
                 },
                {'data': 'interest', 'name': 'interest', 'label': 'Slug', '_unique': True,
                 'className': 'field_req',
                 },
                {'data': 'public', 'name': 'public', 'label': 'Public',
                 '_treatment': {'boolean': {'formfield': 'public', 'dbfield': 'public'}},
                 'ed': {'def': 'yes'},
                 },
                {'data': 'applications', 'name': 'applications', 'label': 'Applications',
                 '_treatment': {'relationship': {'fieldmodel': Application, 'labelfield': 'application',
                                                 'formfield': 'applications', 'dbfield': 'applications',
                                                 'uselist': True}}
                 },
                {'data': 'users', 'name': 'users', 'label': 'Users',
                 '_treatment': {'relationship': {'fieldmodel': User, 'labelfield': 'email',
                                                 'formfield': 'users', 'dbfield': 'users',
                                                 'uselist': True}}
                 },
            ],
            validate=interest_validate,
            servercolumns=None,  # not server side
            idSrc='rowid',
            buttons=['create', 'editRefresh', 'remove'],
            dtoptions={
                'scrollCollapse': True,
                'scrollX': True,
                'scrollXInner': "100%",
                'scrollY': True,
            },
        )
        args.update(kwargs)
        super().__init__(**args)

##########################################################################################
# applications endpoint
###########################################################################################

application_dbattrs = 'id,application'.split(',')
application_formfields = 'rowid,application'.split(',')
application_dbmapping = dict(zip(application_dbattrs, application_formfields))
application_formmapping = dict(zip(application_formfields, application_dbattrs))

def application_validate(action, formdata):
    results = []

    for field in ['application']:
        if formdata[field] and not slug(formdata[field]):
            results.append({ 'name' : field, 'status' : 'invalid slug: must be only alpha, numeral, hyphen' })

    return results

application = DbCrudApiRolePermissions(
                    app = bp,   # use blueprint instead of app
                    db = db,
                    model = Application,
                    version_id_col = 'version_id',  # optimistic concurrency control
                    applications_accepted = 'super-admin',
                    template = 'datatables.jinja2',
                    pagename = 'applications', 
                    endpoint = 'userrole.applications',
                    rule = '/applications',
                    dbmapping = application_dbmapping, 
                    formmapping = application_formmapping, 
                    clientcolumns = [
                        { 'data': 'application', 'name': 'application', 'label': 'Application', '_unique': True,
                          'className': 'field_req',
                          },
                    ],
                    validate = application_validate,
                    servercolumns = None,  # not server side
                    idSrc = 'rowid', 
                    buttons = ['create', 'editRefresh', 'remove'],
                    dtoptions = {
                                        'scrollCollapse': True,
                                        'scrollX': True,
                                        'scrollXInner': "100%",
                                        'scrollY': True,
                                  },
                    )
application.register()




#!/usr/bin/python
###########################################################################################
#   wxextensions - extensions for wxPython widgets
#
#   Date        Author      Reason
#   ----        ------      ------
#   12/07/12    Lou King    Create
#   12/14/12    Lou King    Add deletecallback argument to AutoTextCtrl
#
#   Copyright 2012 Lou King
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###########################################################################################
'''
wxextensions - extensions for wxPython widgets
========================================================
'''

# standard
import optparse

# pypi

# github

# other
import wx   # https://wxpython.org/pages/downloads/ - 2.9 minimum

# home grown

class InvalidParameter(Exception): pass

########################################################################
class AutoTextCtrl(wx.TextCtrl):
# some code copied from wxPython 2.9.4 demo for ListBox
########################################################################
    '''
    TextCtrl enhanced by associated list containing suggestions while typing

    :param *args: :class:`wx.TextCtrl` parameters
    :param **kwargs: :class:`wx.TextCtrl` keyword parameters
    :param items: list of initial items suggested while typing, default []
    :param delcallback: function to call when item deleted from items, default None - takes one parameter, text from deleted item
    '''
    #----------------------------------------------------------------------
    def __init__(self, *args, **kwargs):
    #----------------------------------------------------------------------
        '''
        remember items, instantiate TextCtrl using supplied parameters
        '''
        items = kwargs.pop('items', [])    # default for items is empty list
        if not isinstance(items, list):
            raise InvalidParameter('items must be list')
        self.setitems(items)
        
        self.setdelcallback(kwargs.pop('delcallback', None))
        
        self.typedText = ''
        self.itemndx = 0

        wx.TextCtrl.__init__(self, *args, **kwargs)
        self.Bind(wx.EVT_CHAR,self._OnKey)
        #self.Bind(wx.EVT_KEY_DOWN,self._OnKey)     # returns uppercase key
        
    #----------------------------------------------------------------------
    def setitems(self, items):
    #----------------------------------------------------------------------
        '''
        update the items used for autocompletion

        :param items: list of items suggested while typing
        '''
        self.items = items
        if '' not in self.items:
            self.items.append('')
        self.items.sort(cmp=lambda x,y: cmp(x.lower(), y.lower()))   # NOTE: '' ends up in 0 position
        
    #----------------------------------------------------------------------
    def setdelcallback(self, delcallback=None):
    #----------------------------------------------------------------------
        '''
        update the delcallback function, which is called when an item is deleted from items (list of suggested choices)

        :param delcallback: delcallback(item) is called when an item is deleted from items list.  None to disable
        '''
        self.delcallback = delcallback
        
    #----------------------------------------------------------------------
    def getitems(self):
    #----------------------------------------------------------------------
        '''
        return items used for autocompletion

        :rtype: list of items suggested while typing
        '''
        return self.items
    
    #----------------------------------------------------------------------
    def additem(self, item):
    #----------------------------------------------------------------------
        '''
        add an item to the list used for autocompletion

        :param item: item to add to list of items suggested while typing
        '''
        if item not in self.items:
            self.items.append(item)
            self.setitems(self.items)
        
    #----------------------------------------------------------------------
    def _FindPrefix(self, prefix):
    #----------------------------------------------------------------------
        '''
        find prefix as first substring of any item in self.items, ignoring case
        
        :param prefix: substring to look for
        '''

        if prefix:
            prefix = prefix.lower()
            length = len(prefix)

            # find first item which matches.  NOTE: self.items is a sorted list
            for i in range(len(self.items)):
                item = self.items[i]
                text = item.lower()

                if text[:length] == prefix:
                    # self.log.WriteText('Prefix %s is found.\n' % prefix)
                    return i

        # self.log.WriteText('Prefix %s is not found.\n' % prefix)
        return -1


    #----------------------------------------------------------------------
    def _OnKey(self, evt):
    #----------------------------------------------------------------------
        key = evt.GetKeyCode()

        # text gets added to typedText
        # search for typedText as initial part of item in self.items
        if key >= 32 and key < 127:     # 127 is DELETE key
            self.typedText = self.typedText + chr(key)
            self.itemndx = self._FindPrefix(self.typedText)
            self._SetSelection(self.itemndx)     # if prefix not found, self.typedText is selected

        # backspace removes one character and backs up
        elif key == wx.WXK_BACK:    
            self.typedText = self.typedText[:-1]

            if not self.typedText:
                self._SetSelection(0)   # 0 is always '' item
            else:
                self.itemndx = self._FindPrefix(self.typedText)
                self._SetSelection(self.itemndx)

        # down clears typedText and goes down an item in items
        elif key == wx.WXK_DOWN:    
            if self.itemndx != -1:
                self.typedText = ''
                self.itemndx +=1
                if self.itemndx >= len(self.items):
                    self.itemndx = 0
                self._SetSelection(self.itemndx)
            
            
        # up clears typedText and goes up an item in items
        elif key == wx.WXK_UP:
            if self.itemndx != -1:
                self.typedText = ''
                self.itemndx -=1
                if self.itemndx == -1:
                    self.itemndx = len(self.items)-1
                self._SetSelection(self.itemndx)
            
        # delete deletes selection (from end of typedText through end of found item)
        elif key == wx.WXK_DELETE:
            self.itemndx = -1
            self._SetSelection(self.itemndx)
            
        # ctrl-x deletes current item from list of items.  item is copied into TheClipboard
        elif key == wx.WXK_CONTROL_X:
            if self.itemndx > 0:        # and item in self.items needs to be in this AutoTextCtrl
                item = self.items.pop(self.itemndx)
                self.itemndx = 0
                self._SetSelection(self.itemndx)
                do = wx.TextDataObject()
                do.SetText(item)
                wx.TheClipboard.SetData(do)
                if self.delcallback:
                    self.delcallback(item)
                        
        # ctrl-v is a paste from TheClipboard
        # need to implement it here because TextCtrl hasn't put the data there yet
        elif key == wx.WXK_CONTROL_V:
            do = wx.TextDataObject()
            if wx.TheClipboard.GetData(do): # puts clipboard into do (yes, non-pythonic syntax)
                self.typedText += do.GetText()
            # else (no valid data in TheClipboard) leave self.typedText alone
            self.itemndx = self._FindPrefix(self.typedText)
            self._SetSelection(self.itemndx)
            
        else:
            self.typedText = ''
            evt.Skip()

    #----------------------------------------------------------------------
    def _SetSelection(self, itemndx):
    #----------------------------------------------------------------------
        '''
        set the value of the TextCtrl to the found item,
        select the portion of the item which was not yet typed
        
        :param itemndx: index into self.items
        '''

        # if typed text not found, just set the value
        if itemndx == -1:
            self.SetValue(self.typedText)
            self.SetInsertionPoint(len(self.typedText))
        else:
            self.SetValue(self.items[itemndx])      # NOTE: EVT_COMMAND_TEXT_UPDATED is generated
                                                    # use self.ChangeValue if that is not desired
            self.SetInsertionPoint(len(self.typedText))
            self.SetSelection(len(self.typedText),len(self.items[itemndx]))
        
########################################################################
class _TestWindow(wx.Frame):
########################################################################
    '''
    Test Window for widgets in this package

    :param parent: parent object for this form
    '''

    BTN_OK = wx.NewId()
    BTN_CNCL = wx.NewId()

    #----------------------------------------------------------------------
    def __init__(self,parent):
    #----------------------------------------------------------------------
        self.debug = False

        self.formname = 'test window'
        wx.Frame.__init__(self, parent, wx.ID_ANY, self.formname)
        self.Bind(wx.EVT_CLOSE, self.onClose)

        self.InitUI()
        self.Centre()
        self.Show()

    #----------------------------------------------------------------------
    def InitUI(self):
    #----------------------------------------------------------------------
        """
        Initialize the form
        """
        self.panel = wx.Panel(self)

        font = wx.SystemSettings_GetFont(wx.SYS_SYSTEM_FONT)
        font.SetPointSize(9)

        self.vbox = wx.BoxSizer(wx.VERTICAL)

        # hbox1 - text entry
        hbox1 = wx.BoxSizer(wx.HORIZONTAL)
        st1 = wx.StaticText(self.panel, label='Enter Text')
        st1.SetFont(font)
        hbox1.Add(st1, proportion=1, flag=wx.RIGHT, border=8)
        self.tc = AutoTextCtrl(self.panel,style=wx.TE_PROCESS_ENTER,items=[
            '124 North Market St, Frederick, MD, USA',
            '748 SW Bay Blvd, Newport, OR, USA',
            '30 Germania St, Boston, MA, USA',
            ])
        self.Bind(wx.EVT_TEXT_ENTER, self.onEnter)
        hbox1.Add(self.tc, proportion=5, border=8)
        self.vbox.Add(hbox1, flag=wx.EXPAND|wx.LEFT|wx.RIGHT|wx.TOP, border=10)

        self.vbox.Add((-1, 10))

        self.panel.SetSizerAndFit(self.vbox)
        self.Fit()

    #----------------------------------------------------------------------
    def onEnter(self, evt):
    #----------------------------------------------------------------------
        """
        Search for address, then display rest of frame
        """
        print('text entered: {0}\n'.format(self.tc.GetValue()))
        self.tc.SetValue('')
        return

    #----------------------------------------------------------------------
    def onClose(self, evt):
    #----------------------------------------------------------------------
        """
        Just close the window without updating the station
        """
        self.tc.Destroy()
        self.Destroy()


#######################################################################
class _MyApp(wx.App):
########################################################################

    #----------------------------------------------------------------------
    def __init__(self):
    #----------------------------------------------------------------------
        """
        return MyApp object
        """
        wx.App.__init__(self, False)
        self.frame = _TestWindow(None)


################################################################################
def test():
################################################################################

    usage = "usage: %prog [options]\n"

    parser = optparse.OptionParser(usage=usage)
    (options, args) = parser.parse_args()

    # start the app
    app = _MyApp()
    app.MainLoop()

# ###############################################################################
# ###############################################################################
if __name__ == "__main__":
    test()




#!/usr/bin/python
###########################################################################################
# applogging - define logging for the application
#
#   Date        Author          Reason
#   ----        ------          ------
#   07/06/18    Lou King        Create (from https://github.com/louking/rrwebapp/blob/master/rrwebapp/applogging.py)
#
#   Copyright 2018 Lou King
#
###########################################################################################
'''
applogging - define logging for the application
================================================
'''
# standard
import logging
from logging.handlers import SMTPHandler
from logging import Formatter
from logging.handlers import TimedRotatingFileHandler
from flask import current_app


# pypi

# homegrown

# ----------------------------------------------------------------------
def setlogging():
    # ----------------------------------------------------------------------

    # this is needed for any INFO or DEBUG logging
    current_app.logger.setLevel(logging.DEBUG)

    # # patch werkzeug logging -- not sure why this is being bypassed in werkzeug._internal._log
    # *** for some reason the following code caued debug pin not to be shown, see https://github.com/louking/rrwebapp/issues/300
    # werkzeug_logger = logging.getLogger('werkzeug')
    # werkzeug_logger.setLevel(logging.INFO)

    # TODO: move this to new module logging, bring in from dispatcher
    # set up logging
    ADMINS = ['lou.king@steeplechasers.org']
    if not current_app.debug:
        mail_handler = SMTPHandler('localhost',
                                   'noreply@steeplechasers.org',
                                   ADMINS, 'members exception encountered')
        if 'LOGGING_LEVEL_MAIL' in current_app.config:
            mailloglevel = current_app.config['LOGGING_LEVEL_MAIL']
        else:
            mailloglevel = logging.ERROR
        mail_handler.setLevel(mailloglevel)
        mail_handler.setFormatter(Formatter('''
        Message type:       %(levelname)s
        Location:           %(pathname)s:%(lineno)d
        Module:             %(module)s
        Function:           %(funcName)s
        Time:               %(asctime)s

        Message:

        %(message)s
        '''))
        current_app.logger.addHandler(mail_handler)
        current_app.config['LOGGING_MAIL_HANDLER'] = mail_handler

        logpath = None
        if 'LOGGING_PATH' in current_app.config:
            logpath = current_app.config['LOGGING_PATH']

        if logpath:
            # file rotates every Monday
            file_handler = TimedRotatingFileHandler(logpath, when='W0', delay=True)
            if 'LOGGING_LEVEL_FILE' in current_app.config:
                fileloglevel = current_app.config['LOGGING_LEVEL_FILE']
            else:
                fileloglevel = logging.WARNING
            file_handler.setLevel(fileloglevel)
            current_app.logger.addHandler(file_handler)
            current_app.config['LOGGING_FILE_HANDLER'] = file_handler

            file_handler.setFormatter(Formatter(
                '%(asctime)s %(levelname)s: %(message)s '
                '[in %(pathname)s:%(lineno)d]'
            ))


